/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;
import org.fortiss.tooling.kernel.utils.EcoreUtils;
import org.fortiss.tooling.kernel.utils.KernelModelElementUtils;

/**
 * 
 * @author eder
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 18FB575985A117D61299ECC358019379
 */
public class AnnotationValueService extends
		EObjectAwareServiceBase<IAnnotationValueProvider<IAnnotatedSpecification>> implements
		IAnnotationValueService {

	/** The compositor extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.base.ui.annotationViewPart";

	/** The compositor configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "annotationViewPart";

	/** The handler class attribute name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "binding";

	/** Cache for lookup of registered value providers */
	Map<Class<?>, List<IAnnotationValueProvider<IAnnotatedSpecification>>> valueProviderCache =
			new HashMap<Class<?>, List<IAnnotationValueProvider<IAnnotatedSpecification>>>();

	/** Cache to lookup {@link AnnotationEntry}s for {@link IModelElement}s. */
	Map<IModelElement, List<AnnotationEntry>> annotationEntryCache =
			new HashMap<IModelElement, List<AnnotationEntry>>();

	/**
	 * 
	 */
	public AnnotationValueService() {
		super();
	}

	/**
	 * <p>
	 * In contrast to the implementation of
	 * {@link EObjectAwareServiceBase#getRegisteredHandlers(Class)}, this implementation of
	 * {@link #getRegisteredHandlers(Class)} returns all {@link IAnnotationValueProvider}s that have
	 * been registered for the given class or any of its super-classes or interfaces.
	 * </p>
	 * <p>
	 * TODO Investigate if this behavior is also (optionally) desirable for other services and
	 * should be merged to {@link EObjectAwareServiceBase}.
	 * </p>
	 */
	@Override
	public List<IAnnotationValueProvider<IAnnotatedSpecification>> getRegisteredHandlers(
			Class<?> modelElementClass) {

		// Check if there is a cached result
		List<IAnnotationValueProvider<IAnnotatedSpecification>> valueProviderList =
				valueProviderCache.get(modelElementClass);
		if(valueProviderList != null) {
			return valueProviderList;
		}

		Set<IAnnotationValueProvider<IAnnotatedSpecification>> valueProviderSet =
				new HashSet<IAnnotationValueProvider<IAnnotatedSpecification>>();

		// Queue to walk super-type hierarchy
		Queue<Class<?>> q = new LinkedList<Class<?>>();
		q.add(modelElementClass);

		while(!q.isEmpty()) {
			Class<?> current = q.poll();

			// Determine registered IAnnotationValueProviders for the current class
			// (nearest match)
			List<IAnnotationValueProvider<IAnnotatedSpecification>> currentHandlers =
					super.getRegisteredHandlers(current);
			if(currentHandlers != null) {
				valueProviderSet.addAll(currentHandlers);
			}

			// Also, investigate super-types
			Class<?> superClass = current.getSuperclass();
			if(superClass != null && superClass != Object.class) {
				q.add(superClass);
			}

			for(Class<?> iface : current.getInterfaces()) {
				q.add(iface);
			}
		}

		valueProviderList =
				new ArrayList<IAnnotationValueProvider<IAnnotatedSpecification>>(valueProviderSet);

		// Cache the result as the lookup that walks the entire super-type hierarchy is really
		// expensive (in particular since super.getRegisteredHandlers(current); also partially walks
		// the super-type hierarchy to perform a nearest match.
		valueProviderCache.put(modelElementClass, valueProviderList);

		return valueProviderList;
	}

	/** {@inheritDoc} */
	@Override
	public List<AnnotationEntry> getValues(final IModelElement element) {

		List<AnnotationEntry> cachedResult = annotationEntryCache.get(element);
		if(cachedResult != null) {
			return cachedResult;
		}

		if(element == null) {
			return Collections.emptyList();
		}

		final List<IAnnotationValueProvider<IAnnotatedSpecification>> registeredHandlers =
				getRegisteredHandlers(element.getClass());
		final List<AnnotationEntry> result = new ArrayList<AnnotationEntry>();
		if(registeredHandlers != null && !registeredHandlers.isEmpty()) {

			KernelModelElementUtils.runAsCommand(element, new Runnable() {

				@Override
				public void run() {
					EObject root = element;
					while(!(root instanceof IProjectRootElement)) {
						root = root.eContainer();
					}
					for(IAnnotationValueProvider<IAnnotatedSpecification> annotationProvider : registeredHandlers) {

						List<? extends IModelElement> allChildren =
								getAllChildren(root, element.getClass(), annotationProvider
										.excludeModelElementTypeFromAnnotatedSpecification());

						for(final IModelElement e : allChildren) {
							IAnnotatedSpecification annotatedSpecification =
									EcoreUtils.pickFirstInstanceOf(
											annotationProvider.getAnnotationClazz(),
											e.getSpecifications());
							if(annotatedSpecification == null) {
								try {
									annotatedSpecification =
											annotationProvider
													.getAnnotatedSpecificationForModelElement(e);
								} catch(Exception ex) {
									System.out
											.println("Error instantating new annotation: createAnnotatedSpecificationForModelElement() failed for " +
													annotationProvider.getAnnotationClazz() + ".");
									return;
								}
							}

							boolean add = true;
							for(AnnotationEntry entry : result) {
								if(entry.getModelElement().equals(e)) {
									entry.addNewSpecification(annotationProvider,
											annotatedSpecification);
									add = false;
								}
							}
							if(add) {
								AnnotationEntry entry = new AnnotationEntry(e);
								entry.addNewSpecification(annotationProvider,
										annotatedSpecification);
								result.add(entry);
							}

						}
					}
				}
			});
		}

		annotationEntryCache.put(element, result);

		return result;
	}

	/**
	 * Returns all children from element of the given type and filters out children of the given
	 * {@code modelElementTypesExcludedFromAnnotatedSpecification} list.
	 */
	private <S, T> List<? extends IModelElement> getAllChildren(EObject element,
			Class<? extends IModelElement> type,
			List<Class<? extends EObject>> modelElementTypesExcludedFromAnnotatedSpecification) {

		List<? extends IModelElement> results = EcoreUtils.getChildrenWithType(element, type);
		if(modelElementTypesExcludedFromAnnotatedSpecification != null) {
			List<IModelElement> toRemove = new ArrayList<IModelElement>();
			for(IModelElement modelElement : results) {
				for(Class<?> c : modelElementTypesExcludedFromAnnotatedSpecification) {
					if(c.isInstance(modelElement)) {
						toRemove.add(modelElement);
						break;
					}
				}
			}

			results.removeAll(toRemove);
		}

		return results;
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return HANDLER_CLASS_ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected boolean requiresUniqueHandler() {
		return false;
	}

}
