/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.valueprovider;

import java.util.Collection;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.EditingSupport;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.view.GenericAnnotationView;
import org.fortiss.tooling.kernel.service.base.IEObjectAware;

/**
 * Value provider interface for {@link IAnnotatedSpecification}s that
 * 
 * <p>
 * <ul>
 * <li>establishes the link between the model (data storage) and the UI (data exchange and
 * conversion)</li>
 * <li>provides related annotation data-type specific operations, e.g. factory methods to create
 * <ul>
 * <li>new annotation instances</li>
 * <li>{@link EditingSupport} UI controls for the generic table-based {@link GenericAnnotationView}</li>
 * </ul>
 * </ul>
 * 
 * Concrete annotations need to provide specific implementations of this interface to perform the
 * data exchange and conversion, and to provide the required factory methods.
 * </p>
 * 
 * <p>
 * The interface / some of its methods take two template parameters
 * 
 * <ul>
 * <li><tt>T</tt>: Meta-model type of the specification, i.e. the particular subclass of
 * {@link IAnnotatedSpecification} supported by the respective implementation of this
 * {@link IAnnotationValueProvider} interface.</li>
 * <li><tt>U</tt>: Data type used in the UI to support the respective annotation. For <tt>U</tt>, a
 * reasonable {@code toString()} implementation must be provided that is used by the framework to
 * render the value of the annotation in text-based editing support UI controls.</li>
 * </ul>
 * </p>
 * 
 * @author eder, diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 8BE6A1ECB76A0AD4141B7EE11CAE21DA
 */
public interface IAnnotationValueProvider<T extends IAnnotatedSpecification> extends
		IEObjectAware<IModelElement> {

	/**
	 * Returns the name of the annotation (e.g., used as column label in the table-based
	 * {@link GenericAnnotationView}).
	 */
	public String getAnnotationName(T specification);

	/**
	 * Sets a new name for the annotation (e.g., used as column label in the table-based
	 * {@link GenericAnnotationView}). The return value indicates if the rename operation was
	 * successful.
	 */
	public boolean setAnnotationName(String name, T specification);

	/** Predicate if the given annotation can be edited. */
	public boolean canEdit(T specification);

	/** Predicate if the given annotation can be edited. */
	public boolean canEdit(T specification, String instanceKey);

	/** Returns the value of the annotation. */
	public <U> U getAnnotationValue(T specification) throws Exception;

	/**
	 * Returns the value of a multi-instance annotation's instance denoted by {@code instanceKey}.
	 */
	public <U> U getAnnotationValue(T specification, String instanceKey) throws Exception;

	/**
	 * Returns the {@link EStructuralFeature} which is the return values of the given
	 * specification. This function is useful, if different features are returned depending on the
	 * model element this specification is attached to. Otherwise, the {@link EStructuralFeature}s
	 * can be obtained via the annotation class type.
	 */
	public EStructuralFeature getEStructuralFeatureOf(T specification) throws Exception;

	/**
	 * Sets a new value for the annotation.
	 * 
	 * @throws Exception
	 *             is thrown if the conversion from the UI internal value representation ({@code U})
	 *             to the respective specialization of {@link IAnnotatedSpecification} has failed,
	 *             or if this operation is not available.
	 */
	public <U> void setAnnotationValue(U value, T specification) throws Exception;

	/**
	 * Sets a new value in the annotation from a {@link String} initializer.
	 * 
	 * @throws Exception
	 *             is thrown if the conversion from {@link String} to the respective specialization
	 *             of {@link IAnnotatedSpecification} has failed, or if this operation is not
	 *             available.
	 */
	public void setAnnotationValue(String value, T specification) throws Exception;

	/**
	 * Sets a new value for a multi-instance annotation's instance denoted by {@code instanceKey}.
	 * 
	 * @throws Exception
	 *             is thrown if the conversion from the UI internal value representation ({@code U})
	 *             to the respective specialization of {@link IAnnotatedSpecification} has failed,
	 *             or if this operation is not available.
	 */
	public <U> void setAnnotationValue(U value, T specification, String instanceKey)
			throws Exception;

	/**
	 * Sets a new value for a multi-instance annotation's instance denoted by {@code instanceKey}
	 * from a {@link String} initializer.
	 * 
	 * @throws Exception
	 *             is thrown if the conversion from {@link String} to the respective specialization
	 *             of {@link IAnnotatedSpecification} has failed, or if this operation is not
	 *             available.
	 */
	public void setAnnotationValue(String value, T specification, String instanceKey)
			throws Exception;

	/**
	 * Creates and returns the {@link EditingSupport} support element for the table-based
	 * {@link GenericAnnotationView}. {@link IAnnotationValueProvider}s for specialized
	 * {@link IAnnotatedSpecification} may choose to provide a dedicated, custom
	 * {@link EditingSupport}. For annotations whose {@link IAnnotationValueProvider}s support
	 * multiple instances, an {@code instanceKey} to identify the corresponding instance of the
	 * {@link IAnnotatedSpecification} needs to be provided.
	 * 
	 * @throws Exception
	 *             If the editing support could not be created for the particular
	 *             {@link IAnnotatedSpecification}.
	 */
	public EditingSupport createEditingSupport(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, String instanceKey,
			EStructuralFeature structuralFeature) throws Exception;

	/**
	 * Factory method that creates a new {@link IAnnotatedSpecification} specialization of type T to
	 * the given model {@code element} and returns a reference to it.
	 * 
	 * In case the annotation already exists in the model, this method does not create a new object,
	 * but returns the existing one.
	 * 
	 * @throws Exception
	 *             is thrown if no existing element could be retrieved, and the creation of a new
	 *             element has failed.
	 */
	public T getAnnotatedSpecificationForModelElement(IModelElement element) throws Exception;

	/**
	 * Returns the concrete specialization {@code T} of {@link IAnnotatedSpecification} supported by
	 * this specialization of {@link IAnnotationValueProvider}.
	 */
	public Class<T> getAnnotationClazz();

	/**
	 * Returns a {@link List} of model element types which will not have the
	 * {@link IAnnotatedSpecification} managed by this {@link IAnnotationValueProvider}, refining
	 * the registration of the respective {@link IAnnotatedSpecification} in the
	 * {@code org.fortiss.tooling.base.ui.annotationViewPart} extension point.
	 */
	public List<Class<? extends EObject>> excludeModelElementTypeFromAnnotatedSpecification();

	/**
	 * Returns the instance keys supported by this {@link IAnnotationValueProvider}.
	 * {@link IAnnotationValueProvider}s can either define a static {@link List}, or enable the
	 * dynamic creation of instance keys using {@link #allowsDynamicAnnotationInstances()}. In the
	 * latter case, this method returns current list of dynamic annotation instances.
	 */
	public Collection<String> getInstanceKeys(T specification);

	/**
	 * <p>
	 * Predicate whether the {@link IAnnotatedSpecification} managed by this
	 * {@link IAnnotationValueProvider} contains a value type allowing the storage of dynamically
	 * created instances of the annotation (like a list of strings). For instance, the
	 * {@link GenericAnnotationView} will in this case create UI controls to enable the creation of
	 * these instances.
	 * </p>
	 * <p>
	 * Static instances can be implemented by overriding
	 * {@link #getInstanceKeys(IAnnotatedSpecification)} and returning the {@link List} of
	 * admissible instance keys.
	 * </p>
	 */
	public boolean allowsDynamicAnnotationInstances();

}
