/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view;

import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeSet;

import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.eclipse.gef.editparts.AbstractGraphicalEditPart;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.part.ViewPart;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.AnnotationEntry;
import org.fortiss.tooling.base.ui.annotation.IAnnotationValueService;
import org.fortiss.tooling.kernel.model.IProjectRootElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.ui.util.SelectionUtils;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * This class provides a limited set of variables and methods to annotation views. Its main purposes
 * are the management of selections in the UI and their mapping to the corresponding model elements,
 * and maintaining knowledge about the available annotation entries (see {@link AnnotationEntry}).
 * 
 * @author eder
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 40FF672DC52B17062D1716341177D76A
 */
public abstract class AnnotationViewPartBase extends ViewPart implements ISelectionListener,
		IAnnotationViewPart {

	/** Currently selected object */
	private IModelElement currentlySelectedObject;

	/** Last selected object */
	private IModelElement lastSelectedObject;

	/**
	 * Root element for which the {@link #changeListener} has been installed to (recursively)
	 * watch for model changes (to trigger an update of this {@link AnnotationViewPartBase}).
	 */
	private EObject watchedElement;

	/**
	 * Cache: root element -> annotation entries. An entry is invalidated when an object is added to
	 * / removed from the model.
	 */
	private Map<EObject, Collection<AnnotationEntry>> lastAnnotationEntriesMap =
			new HashMap<EObject, Collection<AnnotationEntry>>();

	/**
	 * {@link Adapter} to watch for the addition, removal or value change of model elements
	 * to/from/in elements visible in this {@link AnnotationViewPartBase} (to trigger according
	 * updates).
	 */
	private final Adapter changeListener = new EContentAdapter() {
		/** {@inheritDoc} */
		@Override
		public void notifyChanged(Notification notification) {
			// Required to actually watch child elements (behavior of EContentAdapter)
			super.notifyChanged(notification);

			// Optimization: Do not not react on the addition of annotations. The latter is actually
			// a consequence of the addition of a new model element for which annotations have
			// been registered.
			//
			// Otherwise, the addition of a new model elements results in a cascade of updates (one
			// for the model element, and one for each of the annotations that have been registered
			// for it)
			if(notification.getEventType() == Notification.ADD &&
					notification.getNewValue() instanceof IAnnotatedSpecification) {
				return;
			}

			if(notification.getEventType() == Notification.ADD ||
					notification.getEventType() == Notification.REMOVE ||
					notification.getEventType() == Notification.SET) {

				Object notifier = notification.getNotifier();
				if((notifier instanceof EObject) &&
						(IPersistencyService.INSTANCE.getTopLevelElementFor((EObject)notifier) != null)) {
					// Ignore update requests for elements that are not hooked to a top-level
					// element,
					// e.g. SET events that are fired when a component is removed (before the actual
					// REMOVE event). Otherwise, the AnnotationValueService would create a NPE
					// during the preparation of the context via
					// KernelModelElementUtils.runAsCommand().

					IModelElement modelElement = null;
					if(notifier instanceof IModelElement) {
						modelElement = (IModelElement)notifier;
					} else if(notifier instanceof IAnnotatedSpecification) {
						modelElement = ((IAnnotatedSpecification)notifier).getSpecificationOf();
					}

					if(modelElement != null) {
						if((notification.getEventType() == Notification.ADD || notification
								.getEventType() == Notification.REMOVE)) {

							// Invalidate cache if elements are added to / removed from the model
							lastAnnotationEntriesMap.put(getRootElement(modelElement), null);
						}

						update(modelElement);
					}
				}
			}
		}
	};

	/** {@inheritDoc} */
	@Override
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		currentlySelectedObject = null;
		currentlySelectedObject = SelectionUtils.checkAndPickFirst(selection, IModelElement.class);

		if(currentlySelectedObject == null) {
			AbstractGraphicalEditPart editPart =
					SelectionUtils.checkAndPickFirst(selection, AbstractGraphicalEditPart.class);
			if(editPart != null && editPart.getModel() instanceof IModelElement) {
				currentlySelectedObject = (IModelElement)editPart.getModel();
			}
		}

		// Optimization: Do nothing if the same object as been selected again
		if(currentlySelectedObject == lastSelectedObject) {
			return;
		}

		if(currentlySelectedObject != null) {
			update(currentlySelectedObject);
			lastSelectedObject = currentlySelectedObject;
		}
	}

	/** Determines root element of the given model element. */
	private EObject getRootElement(IModelElement modelElement) {
		EObject rootElement = modelElement;
		while(!(rootElement instanceof IProjectRootElement)) {
			rootElement = rootElement.eContainer();
		}

		return rootElement;
	}

	/**
	 * Update the {@link AnnotationViewPartBase}: delegate update of concrete view to
	 * {@link #update(Collection)}, and update model change listeners.
	 */
	protected void update(IModelElement modelElement) {
		EObject rootElement = getRootElement(modelElement);

		// Check cache if annotation entries for the given root element have already been determined
		Collection<AnnotationEntry> annotationEntries = lastAnnotationEntriesMap.get(rootElement);

		if(annotationEntries == null) {
			annotationEntries = new TreeSet<AnnotationEntry>(new Comparator<AnnotationEntry>() {
				/** Compare AnnotationEntries by the model element they refer to. */
				@Override
				public int compare(AnnotationEntry a1, AnnotationEntry a2) {
					// Equality of model element implies equal hashCode(), order is arbitrary
					// but not relevant for ensuring uniqueness of annotation entries.
					//
					// int Integer.compare(int x, int y) requires JDK 1.7 which is not available
					// on the build server
					return new Integer(a1.getModelElement().hashCode()).compareTo(a2
							.getModelElement().hashCode());
				}
			});

			// Collect all AnnotationEntries for the current root element and its offspring
			if(rootElement instanceof IModelElement) {
				annotationEntries.addAll(IAnnotationValueService.INSTANCE
						.getValues((IModelElement)rootElement));
			}
			for(IModelElement childModelElement : EcoreUtils.getChildrenWithType(rootElement,
					IModelElement.class)) {
				annotationEntries.addAll(IAnnotationValueService.INSTANCE
						.getValues(childModelElement));
			}

			// Cache annotation entries for the given root element
			lastAnnotationEntriesMap.put(rootElement, annotationEntries);
		}

		// If required, update change listener registration
		if(watchedElement != rootElement) {
			if(watchedElement != null) {
				watchedElement.eAdapters().remove(changeListener);
			}
			watchedElement = rootElement;
			watchedElement.eAdapters().add(changeListener);
		}

		// Update the view
		update(annotationEntries);

	}

	/** Update concrete view */
	protected abstract void update(Collection<AnnotationEntry> annotationEntries);

	/** {@inheritDoc} */
	@Override
	public void createPartControl(Composite parent) {
		getSite().getWorkbenchWindow().getSelectionService().addSelectionListener(this);
	}

	/** {@inheritDoc} */
	@Override
	public IModelElement getCurrentlySelectedObject() {
		return currentlySelectedObject;
	}

	/** {@inheritDoc} */
	@Override
	public void dispose() {
		// Unregister listeners
		getSite().getWorkbenchWindow().getSelectionService().removeSelectionListener(this);
		if(watchedElement != null) {
			watchedElement.eAdapters().remove(changeListener);
			watchedElement = null;
		}

		super.dispose();
	}
}
