/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.compose;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.base.ConnectionSegmentBase;
import org.fortiss.tooling.base.model.base.ConnectorBase;
import org.fortiss.tooling.base.model.base.EntryConnectorBase;
import org.fortiss.tooling.base.model.base.ExitConnectorBase;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.ui.dnd.gef.ConnectionDragContext;
import org.fortiss.tooling.base.ui.editpart.FreeConnectorEditPartBase;
import org.fortiss.tooling.kernel.extension.data.IConnectionCompositionContext;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * Base compositor for connections supporting entry and exit connectors.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: DF00F62A83EC4DC1E62646EA0FB82051
 */
public abstract class ConnectorConnectionCompositorBase<HE extends IHierarchicElement, S extends IConnector, T extends IConnector>
		extends ConnectionCompositorBase<HE, S, T> {

	/** {@inheritDoc} */
	@Override
	public boolean canConnect(S source, T target, HE parent, IConnectionCompositionContext context) {
		// Deny creation of feedback connections in internal view
		if(!allowInternalFeedback() && context instanceof ConnectionDragContext) {
			ConnectionDragContext dc = (ConnectionDragContext)context;
			//
			if(dc.getSource() instanceof FreeConnectorEditPartBase &&
					dc.getTarget() instanceof FreeConnectorEditPartBase) {
				return false;
			}
		}

		if(source instanceof EntryConnectorBase) {
			if(target instanceof EntryConnectorBase) {
				return checkEntryEntry((EntryConnectorBase)source, (EntryConnectorBase)target, null);
			} else if(target instanceof ExitConnectorBase) {
				// use swapping here
				return checkExitEntry((ExitConnectorBase)target, (EntryConnectorBase)source, null);
			}
		} else if(source instanceof ExitConnectorBase) {
			if(target instanceof EntryConnectorBase) {
				return checkExitEntry((ExitConnectorBase)source, (EntryConnectorBase)target, null);
			} else if(target instanceof ExitConnectorBase) {
				return checkExitExit((ExitConnectorBase)source, (ExitConnectorBase)target, null);
			}
		}
		return false;
	}

	/** Checks entry to entry connection. */
	private boolean checkEntryEntry(EntryConnectorBase source, EntryConnectorBase target,
			EObject connection) {
		boolean swapped = false;

		// child input to super input => swap input
		if(source.eContainer().eContainer() == target.eContainer()) {
			EntryConnectorBase temp = source;
			source = target;
			target = temp;
			swapped = true;
		}

		// super input to child input
		if(swapped || source.eContainer() == target.eContainer().eContainer()) {
			return checkSourceOut(source, connection) && checkTargetIn(target, connection);
		}
		return false;
	}

	/** Checks exit to entry connection. */
	private boolean checkExitEntry(ExitConnectorBase source, EntryConnectorBase target,
			EObject connection) {
		if(source.eContainer().eContainer() == target.eContainer().eContainer()) {
			return checkSourceOut(source, connection) && checkTargetIn(target, connection);
		}
		return false;
	}

	/**
	 * Checks exit to exit connection.
	 */
	private boolean checkExitExit(ExitConnectorBase source, ExitConnectorBase target,
			EObject connection) {
		boolean swapped = false;
		if(source.eContainer() == target.eContainer().eContainer()) {
			ExitConnectorBase temp = source;
			source = target;
			target = temp;
			swapped = true;
		}
		if(swapped || source.eContainer().eContainer() == target.eContainer()) {
			return checkSourceOut(source, connection) && checkTargetIn(target, connection);
		}
		return false;
	}

	/**
	 * Checks outgoing connection of source connector.
	 */
	private boolean checkSourceOut(ConnectorBase source, EObject connection) {
		// source has no outgoing connection, is already connected, or allows
		// multiple connections
		return source.getOutgoing().size() == 0 ||
				(connection instanceof ConnectionSegmentBase && ((ConnectionSegmentBase)connection)
						.getSource() == source) || allowOneToMany();
	}

	/** Checks incoming connections of target port. */
	private boolean checkTargetIn(ConnectorBase target, EObject connection) {
		// target has no incoming connection, is already connected, or allows
		// multiple connections.
		return target.getIncoming().size() == 0 ||
				(connection instanceof ConnectionSegmentBase && ((ConnectionSegmentBase)connection)
						.getTarget() == target) || allowManyToOne();
	}

	/** {@inheritDoc} */
	@Override
	public boolean connect(S source, T target, HE parent, IConnectionCompositionContext context) {
		ITopLevelElement modelContext = IPersistencyService.INSTANCE.getTopLevelElementFor(source);
		createConnection(source, target, parent, modelContext);

		return true;
	}

	/**
	 * Configures multiple connection behavior for 1-to-many connections. The
	 * default is allow.
	 */
	protected boolean allowOneToMany() {
		return true;
	}

	/**
	 * Configures multiple connection behavior for many-to-1 connections. The
	 * default is allow.
	 */
	protected boolean allowManyToOne() {
		return true;
	}

	/** Allows internal feedback connection. The default is disallow. */
	protected boolean allowInternalFeedback() {
		return false;
	}
}
